/*
 * Copyright by Intland Software
 *
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Intland Software. ("Confidential Information"). You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Intland.
 */
package com.intland.codebeamer.dbcopy;

import java.sql.Blob;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Class reads the current record from a RecordSet, and does 'something' with that data.
 *
 * @author <a href="mailto:zoltan.luspai@intland.com">Zoltan Luspai</a>
 * @version $Id$
 */
abstract class AbstractRecordReader {
	//private final static Logger logger = Logger.getLogger(AbstractRecordReader.class);

	protected final List columns;
	protected String tableName;

	// map of column types to use: contains upper(columnName) -> (Integer) JDBCType mapping
	protected Map/*<String,Integer>*/ destColumnTypes;
	// the read object values for each column, in same order as columns
	List readObjects;
	// the source database name (see JDBCUtils constants)
	String srcDbName;

	/**
	 * Constructor.
	 * @param columns The names of columns <Strings>
	 * @param tableName The name of the table for logging.
	 * @param columnTypes The map contains (String)columnName->(Integer)jdbcType mapping for destination
	 * @param srcDbName the source database name
	 */
	public AbstractRecordReader (List columns, String tableName, Map destColumnTypes, String srcDbName) {
		this.columns = columns;
		this.tableName = tableName;
		this.destColumnTypes = destColumnTypes;
		this.srcDbName = srcDbName;
	}

	/**
	 * Get the report about the last record's data if an exception occured.
	 * @return The record data
	 */
	public StringBuffer getRecordLog() {
		return buildRecordLog(columns, readObjects);
	}

	/**
	 * Build the record log
	 * @param columns The list of column names
	 * @param readObjects The list of read objects
	 * @return The record log
	 */
	private static StringBuffer buildRecordLog(List columns, List readObjects) {
		final StringBuffer recordLog = new StringBuffer(10000);

		Object readObject;
		// in case of an exception build the record log
		// collect the info of the record to a StringBuffer for logging
		recordLog.append("<record>\n");
		for (int i=0; i< readObjects.size(); i++) {
			readObject = readObjects.get(i);
			recordLog.append(columns.get(i)).append("=");
			if (readObject instanceof byte[]) {
				recordLog.append("byte[], size=" + ((byte[]) readObject).length);
			} else if (!(readObject instanceof java.sql.Blob || readObject instanceof java.sql.Clob)) {
				try {
					recordLog.append(readObject);
				} catch (Exception ignored) {
					// the object may not be read again, ignore this
					recordLog.append("Unknown: " + ignored.getMessage());
				}
			} else {
				recordLog.append("<blob>");
			}
			recordLog.append("\n");
		}
		recordLog.append("</record>");

		return recordLog;
	}

	/**
	 * Copy one record (the current) from the resultset to the prepared writeStmt
	 * @param recordLog The log bufffer contains the record data
	 * @param columns
	 * @param readDataResultSet
	 */
	protected void readOneRecord(ResultSet readDataResultSet) throws SQLException {
		readObjects = new ArrayList(columns.size());
		Object readObject;

		ResultSetMetaData metadata = readDataResultSet.getMetaData();

		for (int i = 1; i <= columns.size(); i++) {
			int jdbcType = metadata.getColumnType(i);
			String columnName = (String) columns.get(i-1);

			int destJdbcType = getDestJdbcType(columnName);
			readObject = readObject(readDataResultSet, columnName, i, jdbcType, destJdbcType);
			readObjects.add(readObject);

			handleColumnValue(readObject, columnName, i, jdbcType, destJdbcType);
		}
	}

	/**
	 * Get the JDBC type for the destination column
	 * @param columnName The name of the column
	 * @return integer jdbc type
	 */
	protected int getDestJdbcType (String columnName) {
		Integer type = (Integer) destColumnTypes.get(columnName.toUpperCase());
		if (type == null) {
			return Types.OTHER;
		}
		return type.intValue();
	}

	// only these fields are read as blobs (byte arrays). Put fields in form of "tablename.fieldname" in here. Compared as ignore-case
	protected static List/*<String>*/ forcedBlobFields = new LinkedList(Arrays.asList(new String[] {
			"forum.attached_document",
			"jtags.str",
			"task_attachment.file_data",
			"object_comment.file_data"
	}));

	/**
	 * Check if the field is forced to read as BLOB
	 * @param fullName the full field name
	 * @return if read as blob
	 */
	protected boolean isForcedBlobField(String fullName) {
		for (Iterator iter = forcedBlobFields.iterator(); iter.hasNext();) {
			String forcedBlobField = (String) iter.next();
			if (fullName.equalsIgnoreCase(forcedBlobField)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Read the column data from the resultset
	 * @param readDataResultSet
	 * @param columnIdx
	 * @param srcJdbcType source jdbc type
	 * @param destJdbcType destination jdbc type
	 * @return The object data
	 * @throws SQLException
	 */
	private Object readObject(ResultSet readDataResultSet, String columnName, int columnIdx, int srcJdbcType, int destJdbcType) throws SQLException {
		// Note: not nice , but only few LONGVARBINARY/BLOB fields contain really binary data, special handling for them!
		String fullName = tableName + "." + columnName;
		if (isForcedBlobField(fullName)) {
			if (JDBCUtils.DBNAME_ORACLE.equals(srcDbName)) {
				// on oracle, read as Blob
				Blob blob = readDataResultSet.getBlob(columnIdx);
				return blob;
			}

			// otherwise and as default read as byte[], it is PostgreSQL compatible
			return readDataResultSet.getBytes(columnIdx);
		}
		// all the rest BLOB/LONGVARBINARY fields are treated as Strings
		if (srcJdbcType == Types.LONGVARBINARY || srcJdbcType == Types.BLOB || srcJdbcType == Types.CLOB) {
			return readDataResultSet.getString(columnIdx);
		}

		if (srcJdbcType == Types.DATE) {
			return readDataResultSet.getDate(columnIdx);
		}
		if (srcJdbcType == Types.TIMESTAMP) {
			return readDataResultSet.getTimestamp(columnIdx);
		}

//		// not used now!
//		if (srcJdbcType == Types.LONGVARBINARY) {
//			// special handling for LONGVARBINARY, sometimes it is a binary data, sometimes it is string
//			// let's decide depending on the destination type
//			if (destJdbcType == Types.VARCHAR) {
//				// read as string
//				return readDataResultSet.getString(columnIdx);
//			}
//
//			Blob blob = readDataResultSet.getBlob(columnIdx);
//			return blob;
//		}
		// default
		return readDataResultSet.getObject(columnIdx);
	}

	/**
	 * Abstract method for subclasses to do something with the just-read column value
	 * @param readObject The read column value
	 * @param columnIdx The index of the column
	 * @param columnName The name of the column
	 * @param jdbcType The original jdbc type of the column
	 * @param destJdbcType The column is read as this jdbc type
	 * @throws SQLException Can throw any SQLException
	 */
	protected abstract void handleColumnValue(Object readObject, String columnName, int columnIdx, int srcJdbcType, int destJdbcType) throws SQLException;

}