/*
 * Copyright by Intland Software
 *
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Intland Software. ("Confidential Information"). You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Intland.
 */
package com.intland.codebeamer.dbcopy;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;

/**
 * Class for that contains the startup configuration of the tool.
 * A method is provided for reading the configuration.
 *
 * @author <a href="mailto:robert.enyedi@intland.com">Robert Enyedi</a>
 * @version $Id$
 */
public class Configuration {

	protected static Logger logger = Logger.getLogger(Configuration.class);

	// property part for commands executed before table copy
	static final String BEFORE_COPY_COMMAND = ".beforeCopyCommand";

	// property part for commands executed after table copy
	static final String AFTER_COPY_COMMAND = ".afterCopyCommand";

	/** The list of table to be copied */
	protected List/*<String>*/ tables;

	/** The connection parameters for the source database */
	protected ConnectionParameters srcConnectionParameters;

	/** The connection parameters for the destination database */
	protected ConnectionParameters destConnectionParameters;

	// if the copying from source to target is forced
	private boolean force = false;

	// if the schema difference is ignored
	private boolean ignoreSchemaDifference = false;

	// configuration properties
	LinkedHashMap properties;

	// if the written records are read back, and verified if they are the same as the original
	protected boolean verify = true;

	// set true if the 1st error should stop the db-copier
	protected boolean FAIL_FAST = false;

	// list of tables with sequences
	private LinkedHashMap/*<String,null*/ sequences;

	// only for testing
	protected Configuration() {
	}

	/**
	 * Constructor.
	 * @param props The configuration properties file.
	 * @param tables The list of table names ordered
	 * @throws IOException if could not read the properties
	 */
	public Configuration(LinkedHashMap props, List tables) throws IOException {
		constructorHelper(props, tables);
	}

	/**
	 * helper method for constructors.
	 * @param props The configuration properties file.
	 * @param tables2 The list of table names ordered
	 * @throws IOException if could not read the properties
	 */
	private void constructorHelper(LinkedHashMap props, List tables2) throws IOException {
		this.properties = props;
		setTables(tables2);
		setSrcConnectionParameters(readConnectionParameters(props, "src"));
		setDestConnectionParameters(readConnectionParameters(props, "dest"));
	}

	/**
	 * Reads the tool configuration from the specified input stream
	 * @param configInput
	 * @param tablesFileInput
	 * @param sequencesFileInput
	 */
	protected Configuration(InputStream configInput, InputStream tablesFileInput, InputStream sequencesFileInput) throws IOException {
		LinkedHashMap props = readAsProperties(configInput);
		LinkedHashMap tables2 = readAsProperties(tablesFileInput);

		// extract table names in order
		List tableNames = getStringKeys(tables2);
		constructorHelper(props, tableNames);

		// read sequences
		sequences = readAsProperties(sequencesFileInput);
	}

	/**
	 * Read the configuration from a set of command line arguments arguments
	 * @param args
	 * @throws IOException
	 * @throws FileNotFoundException
	 * @throws IllegalArgumentException If any arguments are incorrect
	 */
	public Configuration(String[] args) throws FileNotFoundException, IOException {
		this(new FileInputStream(args[0]), new FileInputStream(args[1]), new FileInputStream(args[2]));

		for (int i=3; i<args.length;i++) {
			if (args[i].equals("--force")) {
				setForce(true);
			} else if (args[i].equals("--ignoreSchemaDifference")) {
				setIgnoreSchemaDifference(true);
			} else if (args[i].equals("--noverify")) {
				verify = false;
			} else if (args[i].equals("--verify")) {
				verify = true;
			} else {
				// invalid parameter, don't do anything, just print usage
				throw new IllegalArgumentException("Invalid parameter:" + args[i]);
			}
		}
	}

	/**
	 * Get the keys in order from the LinkedHashMap
	 * @param map The map
	 * @return List contains the keys.
	 */
	private static List getStringKeys(LinkedHashMap map) {
		List result = new LinkedList();
		for (Iterator iter = map.keySet().iterator(); iter.hasNext();) {
			Object key = iter.next();
			result.add(key);
		}
		return result;
	}

	/**
	 * Reads the a properties file, to a LinkedHashMap so it will contain
	 * property=value pairs as keys and values. Lines starting with # will be ignored
	 * This is used to maintain order of the properties as in the file
	 */
	protected static LinkedHashMap/*<String,String>*/ readAsProperties(InputStream input) throws IOException {
		BufferedReader reader = new BufferedReader(new InputStreamReader(input));
		LinkedHashMap map = new LinkedHashMap();

		while(true) {
			String crtLine = reader.readLine();

			if(crtLine == null) {
				return map;
			}

			// Ignore line comments
			if(!crtLine.trim().startsWith("#")) {
				int idx = crtLine.indexOf("=");
				String property;
				String value;
				if (idx == -1) {
					property = crtLine;
					value = null;
				} else {
					property = crtLine.substring(0, idx);
					value = crtLine.substring(idx+1);
				}
				map.put(property, value);
			}
		}
	}

	protected static void checkPropertyPresence(String propertyName, String propertyValue) throws IOException {
		if(propertyValue == null) {
			logger.fatal("Invalid configuration file. Cannot read property '" + propertyName + "'");
			throw new IOException();
		}
	}

	/**
	 * Reads the SQL commands list that are to be executed before or after the copying process
	 */
	private List/*<String>*/ getCommands(String commandKeyPrefix) {

		List commands = new ArrayList();

		for(Iterator it = this.properties.keySet().iterator(); it.hasNext(); ) {
			String crtKey = (String) it.next();

			if(crtKey.startsWith(commandKeyPrefix)) {
				commands.add(properties.get(crtKey));
			}
		}

		return commands;
	}

	/** Reads the connection parameters for the source or destination databases (based on the property key prefix)
	 * @throws IOException */
	protected static ConnectionParameters readConnectionParameters(LinkedHashMap props, String prefix) throws IOException {
		ConnectionParameters connectionParams = new ConnectionParameters();

		connectionParams.setDriver((String) props.get(prefix + ".JDBC.Driver"));
		checkPropertyPresence(connectionParams.getDriver(), prefix + ".JDBC.Driver");

		connectionParams.setUrl((String) props.get(prefix + ".JDBC.ConnectionURL"));
		checkPropertyPresence(connectionParams.getUrl(), prefix + ".JDBC.ConnectionURL");

		connectionParams.setUsername((String) props.get(prefix + ".username"));
		checkPropertyPresence(connectionParams.getUsername(), prefix + ".username");

		connectionParams.setPassword((String) props.get(prefix + ".password"));
		checkPropertyPresence(connectionParams.getPassword(), prefix + ".password");

		// Check for driver parameters
		for(Iterator it = props.keySet().iterator(); it.hasNext(); ) {
			String crtKey = (String) it.next();

			String driverParamsPrefix = prefix + ".Driver.";

			if(crtKey.startsWith(driverParamsPrefix)) {
				connectionParams.getDriverProperties().put(crtKey.substring(driverParamsPrefix.length()), props.get(crtKey));
			}
		}

		return connectionParams;
	}

	public List getAfterCopyCommands(String dbName) {
		return getCommands(dbName +AFTER_COPY_COMMAND);
	}

	public List getBeforeCopyCommands(String dbName) {
		return getCommands(dbName +BEFORE_COPY_COMMAND);
	}

	public ConnectionParameters getDestConnectionParameters() {
		return destConnectionParameters;
	}

	public void setDestConnectionParameters(ConnectionParameters destConnectionParameters) {
		this.destConnectionParameters = destConnectionParameters;
	}

	public ConnectionParameters getSrcConnectionParameters() {
		return srcConnectionParameters;
	}

	public void setSrcConnectionParameters(ConnectionParameters srcConnectionParameters) {
		this.srcConnectionParameters = srcConnectionParameters;
	}

	public List/*<String>*/ getTables() {
		return tables;
	}

	public void setTables(List/*<String>*/ tables) {
		this.tables = tables;
	}

	// if the copying is forced
	public boolean isForce() {
		return force;
	}

	public void setForce(boolean force) {
		this.force = force;
	}

	// if the schema difference is ignored
	public boolean isIgnoreSchemaDifference() {
		return ignoreSchemaDifference;
	}

	public void setIgnoreSchemaDifference(boolean b) {
		ignoreSchemaDifference = b;
	}

	// if the inserted records are read back to verify them
	public boolean isVerify() {
		return verify;
	}

	/**
	 * Determine if the table has sequences.
	 * @param tableName The table'name
	 * @return If table has its own sequence
	 */
	public boolean hasSequence(String tableName) {
		return sequences.containsKey(tableName.toLowerCase());
	}

}
