/*
 * Copyright by Intland Software
 *
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Intland Software. ("Confidential Information"). You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Intland.
 */
package com.intland.codebeamer.dbcopy;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Properties;

import org.apache.log4j.Logger;

/**
 * CodeBeamer database copier tool.
 *
 * @author <a href="mailto:robert.enyedi@intland.com">Robert Enyedi</a>
 * @version $Id$
 */
public class DatabaseCopier {

	protected static Logger logger = Logger.getLogger(DatabaseCopier.class);

	static {
		Locale.setDefault(Locale.US);

		// Init log4j
		//URL log4jprops = DatabaseCopier.class.getResource("./bin/log4j.properties");
		//PropertyConfigurator.configure (log4jprops);
	}

	protected Connection srcConn;
	protected Connection destConn;
	protected Configuration config;

	// list of column names to quote in inserts/selects
	private static List/*<String>*/ columnsToQuote = new LinkedList();

	// if the copy operation as overall was successful
	private boolean success = true;

	// if batch commit is used
	final static boolean BATCH_COMMIT = true;

	// if read-only, nothing is written/deleted/modified in/to the target database
	final static boolean READ_ONLY = false;

	// if after each table is copied do a reconnect, because MySQL does not like too long connections, and may time out
	private static final boolean RECONNECT_AFTER_EACH_TABLE = true;

	// summary report buffer
	private StringBuffer summaryReport = new StringBuffer(
			padStr("Table",30,true)+"\t" + padStr("Source rows", 8, false)
					+ "\t" + padStr("Dest. rows", 8, false) + "\t" + padStr("Next id", 8, false) +"\tSuccessful?\n");

	/**
	 * Constructor
	 * @param config
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws ClassNotFoundException
	 * @throws SQLException
	 */
	public DatabaseCopier(Configuration config) throws InstantiationException, IllegalAccessException, ClassNotFoundException, SQLException {
		this.config = config;

		reconnect();

		// for Oracle and some other databases there are served words used as columns
		// to fix these we'll note the columns needs to be 'quoted'
		String destDBName = JDBCUtils.getDBName(destConn);
		if (destDBName.equals("PostgreSQL")) {
			columnsToQuote.add("offset");
		}

		// ensure that the database schema versions are the same
		ensureDatabaseVersionsSame();

		// check if the target db is empty to avoid data loss
		safetyCheck();
	}

	/**
	 * Connect or reconnect to the source and destination databases.
	 * Needed, because MySQL may occasionally time out long connections.
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 */
	private void reconnect() throws SQLException, InstantiationException, IllegalAccessException, ClassNotFoundException {
		if (srcConn != null) {
			srcConn.commit();	// fix for: org.apache.derby.client.am.SqlException: Cannot close a connection while a transaction is still active.
			srcConn.close();
			srcConn = null;
		}
		if (destConn != null) {
			destConn.commit();
			destConn.close();
			destConn = null;
		}
		srcConn = getDatabaseConnection(this.config.getSrcConnectionParameters());
		destConn = getDatabaseConnection(this.config.getDestConnectionParameters());

		checkDbAndDriverVersion("Source-" , srcConn);
		checkDbAndDriverVersion("Destination-" , destConn);
	}

	/**
	 * Check db and driver version
	 * @param prefix The prefix string for logs
	 * @param conn
	 * @throws SQLException
	 */
	private void checkDbAndDriverVersion(String prefix, Connection conn) throws SQLException {
		DatabaseMetaData metaData = conn.getMetaData();
		String connectionURL = metaData.getURL();
		logger.info(prefix + "Database: " + metaData.getDatabaseProductName() + "-" + metaData.getDatabaseProductVersion());
		logger.info(prefix + "Driver: " + metaData.getDriverName() + " " + metaData.getDriverVersion());
		logger.info(prefix + "URL: <" + connectionURL  + ">");

		// assert driver version
		if (metaData.getDriverName().startsWith("Oracle")) {
			String allowedVersion = "10.2.0.3.0";
			if (! allowedVersion.equals(metaData.getDriverVersion())) {
				String javaHome = System.getProperty("java.home");
				String sep = System.getProperty("file.separator");
				String extDir = javaHome + sep + "lib" + sep + "ext" + sep;
				String msg = "Only version of " + allowedVersion + " is supported for " + metaData.getDriverName();
				msg += "\nCurrent version is: " + metaData.getDriverVersion();
				msg += "\nPlease verify your classpath, including " + extDir;
				msg += " directory, and remove other Oracle jdbc driver jars (ojdbcXXX.jar) from classpath!";
				throw new RuntimeException(msg);
			}
		}
	}

	/**
	 * Ensure that the source database is not empty and the target database is empty.
	 * If they are not so; a warning is printed, and execution is only continued if --force parameter is present.
	 * @throws SQLException
	 * @throws an Exception if the safety check fails
	 */
	private void safetyCheck() throws SQLException {
		boolean srcEmpty = isDatabaseEmpty(srcConn);
		boolean destEmpty = isDatabaseEmpty(destConn);
		if (srcEmpty || !destEmpty) {
			String reason = "";
			if (srcEmpty) {
				reason += "\t- Your source database is EMPTY (why copying it?) !\n";
			}
			if (!destEmpty) {
				reason +="\t- Your destination database is NOT EMPTY!\n";
			}

			if (config.isForce()) {
				System.err.println("Warning! Safety check failed, because:\n" + reason);
				System.err.println("Continuing, because --force is used...");
			} else {
				System.err.println("Error! Safety check failed, because:\n" + reason);
				System.err.println("Please double check that your Source and Destination database connection is properly configured!");
				System.err.println("The execution is stopped to avoid possible data loss (because ALL DATA from the destination database would be deleted)!");
				throw new RuntimeException("Execution stopped, because safety check failed!");
			}
		}
	}

	/**
	 * Ensure that database versions are same
	 * @throw an RuntimeException if the db versions differ
	 */
	private void ensureDatabaseVersionsSame() throws SQLException {
		String[] srcVersion = querySchemaVersion(srcConn);
		String[] destVersion = querySchemaVersion(destConn);
		logger.info("Database schema versions: Source=" +  srcVersion[0] + srcVersion[1] + ", Destination=" + destVersion[0] + destVersion[1]);
		if (!(srcVersion[0].equals(destVersion[0]) && srcVersion[1].equals(destVersion[1]))) {
			if (config.isIgnoreSchemaDifference()) {
				System.err.println("WARNING: the schema version of the source and destination database differs!");
			} else {
				System.err.println("Can not copy database data, because the schema versions differ!");
				System.err.println("Please upgrade database schemas!");
			}
			System.err.println("Source database schema version:" + srcVersion[0] + srcVersion[1]);
			System.err.println("Destination database schema version:" + destVersion[0] + destVersion[1]);
			if (config.isIgnoreSchemaDifference()) {
				System.err.println("Continuing, because --force is used...");
			}
			else {
				throw new RuntimeException("Can not continue, schema versions differ!");
			}
		}
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		logger.info("DatabaseCopier version: $Revision$");
		if(args.length < 3) {
			printUsage();
		} else {
			try {
				Configuration config = null;
				try {
					config = new Configuration(args);
				} catch (IllegalArgumentException ex) {
					printUsage();
					System.exit(1);
				}

				logger.info("Source database URL: " + config.getSrcConnectionParameters().getUrl());
				logger.info("Destination database URL: " + config.getDestConnectionParameters().getUrl());

				DatabaseCopier copier = new DatabaseCopier(config);

				long startTime = System.currentTimeMillis();

				copier.copyTables();

				long endTime = System.currentTimeMillis();

				logger.info("Operation took " + (endTime - startTime)/1000 + " seconds");

				logger.info("\n"+copier.summaryReport);
				if (copier.success) {
					logger.info("Copy finished successfully!");
				} else {
					logger.warn("Copy failed with errors, please verify logs!");
				}
			} catch (Exception ex) {
				logger.fatal("Could not perform the database copy", ex);
			}
		}
	}

	protected static void printUsage() {
		System.out.println("CodeBeamer Database Copier Tool ($Revision$)");
		System.out.println("Usage:");
		System.out.println("\tjava " + DatabaseCopier.class.getName() + " configFile tablesFile sequencesFile [--force] [--ignoreSchemaDifference]");
		System.out.println("Where:");
		System.out.println("\tconfigFile - the configuration file for the tool");
		System.out.println("\ttablesFile - the file containing the list of the tables to be copied");
		System.out.println("\t--force \t- optional parameter, forces execution by skip safety check. Use with extra care!");
		System.out.println("\t--ignoreSchemaDifference - optional parameter, skips schema difference check.");
		System.out.println("\t--noverify - optional parameter, disable verification of the inserted records by reading them back from target database.");
	}

	/**
	 * Performs the copy of all the tables specified in the configuration.
	 * Before and after the copying process, a set of SQL commands are executed as specified in the configuration.
	 */
	public void copyTables() throws SQLException {
		// before and after update commands depend on the database-name (type; i.e. if it is Oracle...)
		String destDBName = JDBCUtils.getDBName(destConn);
		logger.info("Executing SQL commands before the copy process");
		try {
			JDBCUtils.execUpdateCommands(destConn, config.getBeforeCopyCommands(destDBName));
		} catch (SQLException ex) {
			logger.error(ex.getMessage());
		}

		logger.info("Copying tables");
		for(Iterator it = config.getTables().iterator(); it.hasNext(); ) {
			String tableName = (String)it.next();
			TableCopier tableCopier = new TableCopier(srcConn, destConn, columnsToQuote, config.verify, config.FAIL_FAST, config.hasSequence(tableName));
			boolean tableCopiedSuccessfully = false;
			try {
				tableCopiedSuccessfully = tableCopier.copyTable(tableName);
			} catch (Exception ex) {
				String msg = "Couldn't copy table: " + tableName;
				logger.warn(msg, ex);
				if (config.FAIL_FAST) {
					throw new RuntimeException(msg, ex);
				}
			}
			summaryReport.append(tableCopier.getReport());
			success = success && tableCopiedSuccessfully;

			// after each table is copied do a reconnect, because MySQL does not like too long connections, and may time out
			if (RECONNECT_AFTER_EACH_TABLE) {
				try {
					reconnect();
				} catch (Exception ex) {
					logger.error("Exception when reconnecting to database", ex);
					throw new SQLException(ex.getMessage());
				}
			}
		}

		logger.info("Executing SQL commands after the copy process");
		try {
			JDBCUtils.execUpdateCommands(destConn, config.getAfterCopyCommands(destDBName));
		} catch (SQLException ex) {
			logger.error(ex.getMessage());
		}

		// finally fixing passwords
		if (config.getTables().contains("users") && !DatabaseCopier.READ_ONLY) {
			PasswordFixer pwFixer = new PasswordFixer();
			logger.info("Converting passwords of users after migration");
			String pwFixerResult = pwFixer.fixPasswords(destConn, querySchemaVersion(destConn));
			summaryReport.append("Password migration:\t" + pwFixerResult);
		}
	}

	/**
	 * Get the column names of a table
	 * @param tableName
	 * @param conn
	 * @throws SQLException
	 */
	protected void printTableColumnMeta(String tableName, Connection conn) throws SQLException {
		ResultSet rs = JDBCUtils.getTableColumnsInfo(tableName, conn);
		logger.info("Table column meta data for '" + tableName +"' :\n" + JDBCUtils.dumpResultSet(rs));
	}


	/**
	 * Pad a string to the given length by adding spaces
	 * @param str The string to padd
	 * @param len The length of the result string
	 * @param addSpacesToEnd If spaces added to the end
	 * @return padded string
	 */
	static String padStr(String str, int len, boolean addSpacesToEnd) {
		StringBuffer buf = new StringBuffer(str);
		while (buf.length() < len) {
			if (addSpacesToEnd) {
				buf.append(" ");
			} else {
				buf.insert(0, " ");
			}
		}
		return buf.toString();
	}

	/** Creates a database connection based on the specified connection parameters. */
	protected static Connection getDatabaseConnection(ConnectionParameters params) throws SQLException, InstantiationException, IllegalAccessException, ClassNotFoundException {
		DriverManager.registerDriver((Driver)Class.forName(params.getDriver()).newInstance());

		Properties props = new Properties();
	    props.put("user", params.getUsername() );
	    props.put("password", params.getPassword());
	    props.putAll(params.getDriverProperties());

	    Connection conn = DriverManager.getConnection(params.getUrl(), props);
	    conn.setAutoCommit(false);
	    return conn;
	}

	/**
	 * Query the schema version from the db/connection.
	 * @param conn The connection
	 * @return The 2 elem String array 1st is the major, 2nd is the minor schema version
	 * @throws SQLException
	 */
	public String[] querySchemaVersion(Connection conn) throws SQLException {
		String[] version = (String[]) JDBCUtils.query(conn, "SELECT major, minor FROM version",
				new JDBCUtils.IResultSetHandler() {

					public Object handleResults(ResultSet rs) throws SQLException {
						rs.next();
						return new String[] {rs.getString(1), rs.getString(2)};
					}}
		);
		return version;
	}

	/**
	 * Check if the database is empty, by checking few key tables.
	 */
	public boolean isDatabaseEmpty(Connection conn) throws SQLException {
		int numExisting = JDBCUtils.querySingleIntValue(conn, "SELECT count(*) FROM existing");
		return (numExisting == 0);
	}

}

