/*
 * Copyright by Intland Software
 *
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Intland Software. ("Confidential Information"). You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Intland.
 */
package com.intland.codebeamer.dbcopy;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

/**
 * Simple JDBC utility methods.
 *
 * @author <a href="mailto:zoltan.luspai@intland.com">Zoltan Luspai</a>
 * @version $Id$
 */
public class JDBCUtils {

	protected static Logger logger = Logger.getLogger(DatabaseCopier.class);

	// Database name for getDbName() for differnt databases
	public static final String DBNAME_ORACLE = "Oracle";
	public static final String DBNAME_DERBY = "Apache Derby";
	public static final String DBNAME_POSTGRE_SQL = "PostgreSQL";
	public static final String DBNAME_MY_SQL = "MySQL";

	/**
	 * Query a single int value as result from a connection
	 * @param conn
	 * @param query
	 * @return The single int value, or -1 if no value
	 * @throws SQLException
	 */
	public static int querySingleIntValue(Connection conn, final String query) throws SQLException {
		IResultSetHandler handler = new IResultSetHandler() {

			public Object handleResults(ResultSet rs) throws SQLException {
				int result = -1;
				if(rs.next()) {
					result = rs.getInt(1);
				} else {
					logger.warn("Cannot retrieve result.");
					logger.debug("Query is: " + query);
				}
				return new Integer(result);
			}
		};

		Integer result = (Integer) query (conn, query, handler);

		return result.intValue();
	}

	/**
	 * Query a single int value as result from a connection
	 * @param conn
	 * @param query
	 * @param handler The callback handler will process the resultset
	 * @throws SQLException
	 */
	public static Object query(Connection conn, String query, IResultSetHandler handler) throws SQLException {
		logger.debug("Executing query:" + query);
		ResultSet rs = null;
		Statement stmt = null;
		try {
			stmt = conn.createStatement();
			rs = stmt.executeQuery(query);
			return handler.handleResults(rs);
		}
		finally {
			if(rs != null) {
				rs.close();
			}

			if(stmt != null) {
				stmt.close();
			}
		}
	}

	// 'Spring' style callback interface for query results
	public static interface IResultSetHandler {

		/**
		 * Handle a Resultset
		 * @param rs
		 * @return Any result value will be returned from query() call
		 * @throws SQLException
		 */
		public Object handleResults(ResultSet rs) throws SQLException;

	}

	/**
	 * Extract one column information as Strings from a Resultset
	 * @param rs
	 * @param columnName
	 * @return The list will contain the values for the column
	 * @throws SQLException
	 */
	public static List extractResultsetColumnAsStrings(ResultSet rs, String columnName) throws SQLException{
		List result = new LinkedList();
		try {
			if(rs == null) {
				logger.fatal("Cannot read information");
				return null;
			}

			while (rs.next()) {
				result.add(rs.getString(columnName));
			}
		} finally {
			if (rs != null) {
				rs.close();
			}
		}
		return result;
	}

	// dump resultset to a String as Strings
	public static String dumpResultSet(ResultSet rs) throws SQLException {
		StringBuffer dump = new StringBuffer();
		ResultSetMetaData metadata = rs.getMetaData();
		try {
			// header
			for (int i=1; i<= metadata.getColumnCount(); i++) {
				if (i != 1) {
					dump.append("\t");
				}
				dump.append(metadata.getColumnLabel(i));
			}
			dump.append("\n");

			// data
			while (rs.next()) {
				for (int i=1; i<= metadata.getColumnCount(); i++) {
					if (i != 1) {
						dump.append("\t");
					}
					dump.append(rs.getString(i));
				}
				dump.append("\n");
			}
		} finally {
			if (rs != null) {
				rs.close();
			}
		}
		return dump.toString();
	}

	/**
	 * Execute a list of SQL INSERT/UPDATE/DELETE commands using the specified database connection.
	 * @param conn The connection
	 * @param commands the list of commands
	 * @return number of records changed
	 */
	public static int execUpdateCommands(Connection con, List commands) throws SQLException {
		Statement stmt = null;
		int total = 0;

		try {
			stmt = con.createStatement();

			for(Iterator it = commands.iterator(); it.hasNext(); ) {

				String crtCommand = (String) it.next();

				if(crtCommand.trim().length() > 0) {
					logger.debug("Executing update command: " + crtCommand);
					total += stmt.executeUpdate(crtCommand);
				}
			}
		} finally {
			stmt.close();
		}
		return total;
	}

	/**
	 * Execute one command without any results expected
	 * @param conn
	 * @param command The command
	 * @return number of records changed
	 * @throws SQLException
	 */
	public static int execUpdateCommand(Connection conn, String command) throws SQLException {
		return execUpdateCommands(conn, Arrays.asList(new String[] {command}));
	}

	/**
	 * Get the database's name
	 * @param conn
	 * @return The db name like "MySQL"
	 * @throws SQLException
	 */
	public static String getDBName(Connection conn) throws SQLException {
		String dbName = conn.getMetaData().getDatabaseProductName();
		return dbName;
	}

	/**
	 * Determine if the database's meta-data functions are case-sensitive.
	 * This means, the table names, etc must be converted to upper-case, when searching for table's columnns, etc.
	 * @param conn The connection
	 * @return If it is case sensitive
	 * @throws SQLException
	 */
	private static boolean isMetaDataFunctionsCaseSensitive(Connection conn) throws SQLException {
		// Oracle & Derby is case sensitive on name
		String dbName = getDBName(conn);
		return dbName.equals(DBNAME_ORACLE) || dbName.equals(DBNAME_DERBY);
	}

	/**
	 * Get the name of the primary key field in a table
	 * @param conn The connection
	 * @param tableName The name of the table
	 * @return The pk field name, null if no primary key, or throws a RuntimeException if more then one primary key columns present!
	 */
	public static String getPrimaryKeyFieldName(Connection conn, String tableName) throws SQLException {
		// Oracle & Derby is case sensitive on name
		if (isMetaDataFunctionsCaseSensitive(conn)) {
			tableName = tableName.toUpperCase();
		}
		ResultSet rs = conn.getMetaData().getPrimaryKeys(null, null, tableName);
		List pkColumns = JDBCUtils.extractResultsetColumnAsStrings(rs, "COLUMN_NAME");
		if (pkColumns.size() > 1) {
			throw new RuntimeException("Multiple keys found for table: " + tableName + ", unable to determine which is auto-incremented!");
		}
		// null if no pk
		if (pkColumns.size() == 0) {
			return null;
		}
		return (String) pkColumns.get(0);
	}

	// contains the jdbc type name Integer->String conversions
	private static Map/*<Integer>,<String>*/ jdbcTypeNames = new HashMap();

	static {
		jdbcTypeNames.put(new Integer(-7),"BIT");
		jdbcTypeNames.put(new Integer(-6),"TINYINT");
		jdbcTypeNames.put(new Integer(5),"SMALLINT");
		jdbcTypeNames.put(new Integer(4),"INTEGER");
		jdbcTypeNames.put(new Integer(-5),"BIGINT");
		jdbcTypeNames.put(new Integer(6),"FLOAT");
		jdbcTypeNames.put(new Integer(7),"REAL");
		jdbcTypeNames.put(new Integer(8),"DOUBLE");
		jdbcTypeNames.put(new Integer(2),"NUMERIC");
		jdbcTypeNames.put(new Integer(3),"DECIMAL");
		jdbcTypeNames.put(new Integer(1),"CHAR");
		jdbcTypeNames.put(new Integer(12),"VARCHAR");
		jdbcTypeNames.put(new Integer(-1),"LONGVARCHAR");
		jdbcTypeNames.put(new Integer(91),"DATE");
		jdbcTypeNames.put(new Integer(92),"TIME");
		jdbcTypeNames.put(new Integer(93),"TIMESTAMP");
		jdbcTypeNames.put(new Integer(-2),"BINARY");
		jdbcTypeNames.put(new Integer(-3),"VARBINARY");
		jdbcTypeNames.put(new Integer(-4),"LONGVARBINARY");
		jdbcTypeNames.put(new Integer(0),"NULL");
		jdbcTypeNames.put(new Integer(1111),"OTHER");
		jdbcTypeNames.put(new Integer(2000),"JAVA_OBJECT");
		jdbcTypeNames.put(new Integer(2001),"DISTINCT");
		jdbcTypeNames.put(new Integer(2002),"STRUCT");
		jdbcTypeNames.put(new Integer(2003),"ARRAY");
		jdbcTypeNames.put(new Integer(2004),"BLOB");
		jdbcTypeNames.put(new Integer(2005),"CLOB");
		jdbcTypeNames.put(new Integer(2006),"REF");
		jdbcTypeNames.put(new Integer(70),"DATALINK");
		jdbcTypeNames.put(new Integer(16), "BOOLEAN");
	}

	// get name for a JDBC-type int
	public static String getJDBCTypeName(int jdbcType) {
		return (String) jdbcTypeNames.get(new Integer(jdbcType));
	}

	/**
	 * Query the db column info resultset from for a table.
	 * @param tableName
	 * @param conn
	 * @return The resultset contains the info
	 * @throws SQLException
	 */
	public static ResultSet getTableColumnsInfo(String tableName, Connection conn) throws SQLException {
		DatabaseMetaData dbMetaData = conn.getMetaData();
		ResultSet rs = null;

		// Oracle is case sensitive on name
		if (isMetaDataFunctionsCaseSensitive(conn)) {
			tableName = tableName.toUpperCase();
		}

		rs = dbMetaData.getColumns(null, null, tableName, null);
		return rs;
	}

	/**
	 * Get the column names of a table.
	 * @param tableName
	 * @param conn
	 * @return List contains the column name strings
	 * @throws SQLException
	 */
	public static List getTableColumns(String tableName, Connection conn) throws SQLException {
		ResultSet rs = getTableColumnsInfo(tableName, conn);

		List columns = JDBCUtils.extractResultsetColumnAsStrings(rs, "COLUMN_NAME");
		return columns;
	}

	/**
	 * Get the jdbc-types for each columns in a table.
	 * Returns a map of <String>ColumnName-><Integer>jdbc-type (java.sql.Types)
	 * @param tableName
	 * @param conn
	 * @return The map
	 * @throws SQLException
	 */
	public static Map getColumnTypes(String tableName, Connection conn) throws SQLException {
		Map result = new HashMap();
		ResultSet rs = getTableColumnsInfo(tableName, conn);

		try {
			while (rs.next()) {
				String columnName = rs.getString("COLUMN_NAME").toUpperCase();
				Integer jdbcType = new Integer(rs.getInt("DATA_TYPE"));
				result.put(columnName, jdbcType);
			}
		} finally {
			rs.close();
		}
		return result;
	}


}
