/*
 * Copyright by Intland Software
 *
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Intland Software. ("Confidential Information"). You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Intland.
 */
package com.intland.codebeamer.dbcopy;

import java.io.ByteArrayInputStream;
import java.sql.Blob;
import java.sql.Clob;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.List;
import java.util.Map;

/**
 * Resultset handler copies current record of ResultSet to an other database.
 *
 * @author <a href="mailto:zoltan.luspai@intland.com">Zoltan Luspai</a>
 * @version $Id$
 */
class RecordCopier extends AbstractRecordReader implements JDBCUtils.IResultSetHandler {
	private final PreparedStatement writeStmt;
	// number of records written in the last insert
	private int writtenRecords;
	// verifies if the inserted record is the same as the original
	private RecordVerifier recordVerifier;

	/**
	 * Constructor.
	 * @param columns The names of columns <Strings>
	 * @param writeStmt Prepared write statements where the data is copied to
	 * @param tableName The name of the table for logging.
	 * @param columnTypes The map contains (String)columnName->(Integer)jdbcType mapping for destination
	 * @param srcDbName the source database name
	 */
	RecordCopier(List columns, PreparedStatement writeStmt, String tableName, Map destColumnTypes, RecordVerifier recordVerifier, String srcDbName) {
		super(columns, tableName, destColumnTypes, srcDbName);
		this.writeStmt = writeStmt;
		this.recordVerifier = recordVerifier;
	}

	/**
	 * Implemented interface method
	 */
	public Object handleResults(ResultSet rs) throws SQLException {
		return new Integer(copyOneRecord(rs));
	}

	/**
	 * Copy one record (the current) from the resultset to the prepared writeStmt
	 * @param writeStmt
	 * @param recordLog The log bufffer contains the record data
	 * @param columns
	 * @param readDataResultSet
	 * @return number of records got copied
	 * @throws SQLException
	 */
	private int copyOneRecord(ResultSet readDataResultSet) throws SQLException {
		super.readOneRecord(readDataResultSet);

		// after the record is read insert it to target record
		if (DatabaseCopier.READ_ONLY) {
			// do not write, but simulate one record
			writtenRecords = 1;
		} else {
			writtenRecords = writeStmt.executeUpdate();
		}

		// verify the recod
		if (recordVerifier != null) {
			recordVerifier.verify(super.readObjects);
		}
		return writtenRecords;
	}

	/**
	 * Implemented abstract record.
	 * @see com.intland.codebeamer.dbcopy.AbstractRecordReader#handleColumnValue(java.lang.Object, java.lang.String, int, int, int)
	 */
	protected void handleColumnValue(Object readObject, String columnName, int columnIdx, int srcJdbcType, int destJdbcType) throws SQLException {
		writeObject(readObject, columnName, columnIdx, srcJdbcType, destJdbcType);
	}

	/**
	 * Write the object as parameter to the writeStmt
	 * @param readObject
	 * @param columnIdx
	 * @param columnName
	 * @param jdbcType
	 * @param destJdbcType
	 * @throws SQLException
	 */
	private void writeObject(Object readObject, String columnName, int columnIdx, int srcJdbcType, int destJdbcType) throws SQLException {
		// special handling for LONGVARBINARY
//		if (jdbcType == Types.LONGVARBINARY) {
//			// MySQL returns a byte[] array for LONGVARBINARY
//			byte[] bytes = (byte[]) readObject;
//			//writeStmt.setBytes(columnIdx, bytes );
//			writeStmt.setBinaryStream(columnIdx, new ByteArrayInputStream(bytes), bytes.length);
//			return;
//		}

		// Special handle for Clob/Blobs handle them as InputStreams
		if(readObject instanceof Clob) {
			Clob clob = (Clob) readObject;
			writeStmt.setAsciiStream(columnIdx, clob.getAsciiStream(), (int) clob.length());
			return;
		}
		if(readObject instanceof Blob) {
			Blob blob = (Blob)readObject;
			writeStmt.setBinaryStream(columnIdx, blob.getBinaryStream(), (int) blob.length());
			return;
		}
		if (readObject instanceof byte[]) {
			byte[] bytes = (byte[]) readObject;
			writeStmt.setBinaryStream(columnIdx, new ByteArrayInputStream(bytes), bytes.length);
			return;
		}

		// when writing a String to a Blob/Clob write as stream too
		// to avoid ORA-01465: invalid hex number error
		if ((destJdbcType == Types.BLOB /* Oracle */ || destJdbcType == Types.BINARY /* PostgreSQL */)
				&& readObject instanceof String) {
			String str = (String ) readObject;
			byte[] bytes = str.getBytes();
			writeStmt.setBinaryStream(columnIdx, new ByteArrayInputStream(bytes), bytes.length);
			return;
		}
//		if (destJdbcType == Types.CLOB && readObject instanceof String) {
//			String str = (String ) readObject;
//			writeStmt.setAsciiStream(columnIdx, new StringInputStream(str), str.length());
//			return;
//		}

		// default, for normal objects, let the jdbc driver do the handling
		writeStmt.setObject(columnIdx, readObject /*, jdbcType // not passed, may cause conversion errors*/);
	}

}

