/*
 * Copyright by Intland Software
 *
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Intland Software. ("Confidential Information"). You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Intland.
 */
package com.intland.codebeamer.dbcopy;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.sql.Blob;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

/**
 * Class verifies that the newly inserted record is the same as the original record.
 *
 * @author <a href="mailto:zoltan.luspai@intland.com">Zoltan Luspai</a>
 * @version $Id$
 */
public class RecordVerifier extends AbstractRecordReader {

	private final static Logger logger = Logger.getLogger(RecordVerifier.class);

	// the statement to verify, 1 parameter, is the primary key
	private PreparedStatement readStmt;
	// name of the PK column
	private String primaryKey;
	// index of the primary key column
	private int pkColumnIdx;
	// if has primary key
	private boolean hasPK = false;

	// statistics about which columns were verified
	private Set/*<String>*/ columnsVerified = new HashSet();

	/**
	 * Constructor.
	 * @param columns
	 * @param tableName
	 * @param destColumnTypes
	 * @param srcDbName The source database name
	 * @throws SQLException
	 */
	public RecordVerifier(List columns, String tableName, Map destColumnTypes, Connection destConn, String fetchOneRecordSQL, String primaryKey, String srcDbName) throws SQLException {
		super(columns, tableName, destColumnTypes, srcDbName);
		this.readStmt = destConn.prepareStatement(fetchOneRecordSQL);
		this.primaryKey = primaryKey;

		// find the pk column's index
		for (int i=0; i< super.columns.size(); i++) {
			String columnName = (String) columns.get(i);
			if (columnName.equalsIgnoreCase(primaryKey)) {	// equalsIgnoreCase is used, because Derby returns PK key as uppercase
				pkColumnIdx = i;
				hasPK = true;
			}
		}
	}

	/* (non-Javadoc)
	 * @see com.intland.codebeamer.dbcopy.AbstractRecordReader#handleColumnValue(java.lang.Object, java.lang.String, int, int, int)
	 */
	protected void handleColumnValue(Object readObject, String columnName, int columnIdx, int srcJdbcType, int destJdbcType) throws SQLException {
		// noop
	}

	/**
	 * Verify that the database contain the same record as the current read object
	 * @param srcReadObjects The objects read from source database
	 * @throws SQLException for any SQL errors
	 */
	public void verify(List srcReadObjects) throws SQLException {
		// if no pk value that's an error!
		if (!hasPK) {
			throw new RuntimeException("Primary key value is empty, tableName=" + tableName + ", primaryKey=" + primaryKey);
		}
		Object pkValue = srcReadObjects.get(pkColumnIdx);

		// special handling for table sym_types table, because its key is padded, so would not find record
		// so we don't verify this
		if (tableName.equalsIgnoreCase("sym_types")) {
			return;
		}

		// read back the record from destination connection
		readStmt.clearParameters();
		readStmt.setObject(1, pkValue);
		ResultSet rs = readStmt.executeQuery();
		try {
			boolean hasNext = rs.next();
			if (! hasNext) {
				throw new RuntimeException("Record not found in destination database, tableName=" + tableName +", primaryKey=" + primaryKey +", keyValue=" + pkValue);
			}
			// read the current record from destination db
			super.readOneRecord(rs);

			// verify that two data sets are the same
			verifyDataEquals(srcReadObjects, this.readObjects);

			hasNext = rs.next();
			if (hasNext) {
				throw new RuntimeException("More than one record matching in destination database, tableName=" + tableName +", primaryKey=" + primaryKey +", keyValue=" + pkValue);
			}
		} finally {
			rs.close();
		}
	}

	// if the same string is ok if only this is padded
	private boolean ALLOW_PADDED_STRINGS = true;

	// full field name in form of "tablename.fieldname" which are allowed as padded strings
	// these are CHAR fields for Oracle, but handled properly in iBatis XMLs
	private static List/*<String>*/ fieldsAllowedPaddedStrings = Arrays.asList(new String[] {
			"users.passwd",
			"sym_types.type",
			"sym_types.descr",
			"object_association_type.description_format"
	});

	// if the empty strings can become null during conversion
	private boolean ALLOW_EMPTY_STRINGS_AS_NULL = true;

	/**
	 * Verify that two data-sets are equal
	 * @param srcReadObjects The objects from the src database
	 * @param destReadObjects The objects from the dest database
	 */
	private void verifyDataEquals(List srcReadObjects, List destReadObjects) {
		for (int i=0; i<srcReadObjects.size(); i++) {
			boolean columnVerified = false;
			Object srcObj = srcReadObjects.get(i);
			Object destObj = destReadObjects.get(i);
			String columnName = (String) columns.get(i);
			String fullFieldName = tableName + "." + columnName;
			boolean different = false;

			// jdbc-type of the destination column
			int destColumnType = getDestJdbcType(columnName);

			if (srcObj == null || destObj == null) {
				// check they are both null
				if (srcObj != null || destObj != null) {
					different = true;
				}
				// in Oracle the empty string (or empty blobs,byte[]) may become null
				try {
					if (different
							&& (srcObj == null || StringUtils.trimToNull("" + srcObj) == null || isEmptyBlob(srcObj))
							&& (destObj == null || StringUtils.trimToNull("" + destObj) == null || isEmptyBlob(destObj))) {
						if (ALLOW_EMPTY_STRINGS_AS_NULL) {
							different = false;
						} else {
							logger.info("Difference in strings, empty string does not equal null on: " + fullFieldName + ", src=<" + srcObj + ">, dest=<" + destObj +">");
						}
					}
				} catch (SQLException ex) {
					logger.warn("Can not check if blob field is empty on " + fullFieldName);
					if (logger.isDebugEnabled()) {
						logger.debug("Exception:" , ex);
					}
				}
			} else {
				// different comparisions for different jdbc-types...
				// TODO: more types, like blob, etc?
				switch (destColumnType) {
				case Types.INTEGER:
				case Types.CHAR:
				case Types.VARCHAR:
					columnVerified = true;
					// compare as toString()
					if (!srcObj.toString().equals(destObj.toString())) {
						different = true;
					}

					// in Oracle strings might be padded
					if (different && StringUtils.trimToEmpty(srcObj.toString()).equals(StringUtils.trimToEmpty(destObj.toString()))) {
						if (ALLOW_PADDED_STRINGS || fieldsAllowedPaddedStrings.contains(fullFieldName.toLowerCase())) {
							different = false;
						} else {
							logger.info("Difference in padding on " + fullFieldName);
						}
					}
					break;
				case Types.TIMESTAMP:
					columnVerified = true;
					different = compareDatesUpTo(srcObj, destObj, Calendar.SECOND);
					break;
				case Types.DATE:
					columnVerified = true;
					different = compareDatesUpTo(srcObj, destObj, Calendar.DAY_OF_MONTH);
					break;
				case Types.DECIMAL:
					columnVerified = true;
					different = compareDouble(srcObj, destObj);
					break;
				default:
					columnVerified = false;
					break;
				}
				// check if two objects are byte[] and compare them
				try {
					byte[] srcBytes = getBlobBytes(srcObj);
					byte[] destBytes = getBlobBytes(destObj);
					if (srcBytes != null && destBytes != null) {
						columnVerified = true;
						different = compareByteArrays(srcBytes, destBytes);
					}
				} catch (SQLException ignored) {
					logger.debug("Ignored-", ignored);
				}
			}
			if (different) {
				// one of the objects are not null
				throw new RuntimeException("Value difference: column = " + columnName +", src value=<" + srcObj +">, dest value=<" + destObj +">, jdbcType=" + JDBCUtils.getJDBCTypeName(destColumnType));
			}
			// maintain statistics
			if (columnVerified && !columnsVerified.contains(columnName)) {
				columnsVerified.add(columnName);
			}
		}
	}

	/**
	 * Compare two byte array
	 * @param srcBytes
	 * @param destBytes
	 * @return true if they are different
	 */
	static boolean compareByteArrays(byte[] srcBytes, byte[] destBytes) {
		boolean different = false;
		if (srcBytes.length != destBytes.length) {
			different = true;
			logger.info("Different length Blobs: src-length:" + srcBytes.length + ", dest-length:" + destBytes.length);
		} else {
			// compare byte array content
			for (int j=0; j<srcBytes.length && !different; j++) {
				if (srcBytes[j] != destBytes[j]) {
					different = true;
					logger.info("Blob difference at position:" + j);
				}
			}
		}
		return different;
	}

	/**
	 * Try to read the object as blob/byte array
	 * @param obj
	 * @return The byte[] if it was a byte[] or a Blob, or null otherwise
	 */
	private byte[] getBlobBytes(Object obj) throws SQLException {
		if (obj instanceof byte[]) {
			byte[] bytea = (byte[]) obj;
			return bytea;
		}
		try {
			if (obj instanceof Blob) {
				Blob blob = (Blob) obj;
				ByteArrayOutputStream buf = new ByteArrayOutputStream();
				streamCopy(blob.getBinaryStream(), buf);
				return buf.toByteArray();
			}
			if (obj instanceof Clob) {
				Clob clob = (Clob) obj;
				ByteArrayOutputStream buf = new ByteArrayOutputStream();
				streamCopy(clob.getAsciiStream(), buf);
				return buf.toByteArray();
			}
		} catch (IOException ioex) {
			logger.warn("Can not read stream", ioex);
			throw new SQLException("Can not read stream: " + ioex.getMessage());
		}
		return null;
	}

	/**
	 * Copy content of an input stream to an output stream
	 * @param input
	 * @param output The output stream, will be closed
	 * @throws IOException
	 */
	private void streamCopy(InputStream input, OutputStream output) throws IOException {
		byte[] buf = new byte[1000];
		long totalBytes = 0;
		for (int bytesRead = 0;
					(bytesRead = input.read(buf, 0, buf.length) ) != -1;
					totalBytes += bytesRead)
		{
			output.write(buf, 0, bytesRead);
		}
		output.flush();
		output.close();
	}

	/**
	 * Decide if the object is an empty blob or empty array (from blob).
	 * @param obj
	 * @return If it is empty
	 * @throws SQLException
	 */
	private boolean isEmptyBlob(Object obj) throws SQLException{
		if (obj instanceof byte[]) {
			byte[] bytea = (byte[]) obj;
			return bytea.length == 0;
		}
		if (obj instanceof Blob) {
			Blob blob = (Blob) obj;
			return blob.length() == 0;
		}
		if (obj instanceof Clob) {
			Clob clob = (Clob) obj;
			return clob.length() == 0;
		}
		return false;
	}

	/**
	 * Compare Decimal/Double values
	 * @param srcObj
	 * @param destObj
	 * @return true if different
	 */
	static boolean compareDouble(Object srcObj, Object destObj) {
		double srcValue = getDouble(srcObj);
		double destValue = getDouble(destObj);
		if (Math.abs(srcValue - destValue)>0.0001) {
			return true;
		}
		return false;
	}

	/**
	 * Get the double value out of object
	 * @param obj
	 * @return The double value
	 */
	static double getDouble(Object obj) {
		if (obj instanceof Number) {
			return ((Number) obj).doubleValue();
		}
		String msg = "Could not get double value from:" + obj +", class:" + obj.getClass().getName();
		logger.info(msg);
		throw new RuntimeException(msg);
	}

	// calendar fields in order
	final static int fields[] = new int[] { Calendar.YEAR, Calendar.MONTH, Calendar.DAY_OF_MONTH, Calendar.HOUR_OF_DAY, Calendar.MINUTE, Calendar.SECOND };

	/**
	 * @param srcObj The source object
	 * @param destObj
	 * @param lastfield The last field to compare to
	 * @return if they are different
	 */
	static boolean compareDatesUpTo(Object srcObj, Object destObj, int lastfield) {
		Date srcDate = getDate(srcObj);
		Date destDate = getDate(destObj);

		Calendar srcCal = Calendar.getInstance();
		srcCal.setTime(srcDate);
		Calendar destCal = Calendar.getInstance();
		destCal.setTime(destDate);

		for (int i=0; i < fields.length; i++) {
			int field = fields[i];
			int srcValue = srcCal.get(field);
			int destValue = destCal.get(field);
			if (srcValue != destValue) {
				// different!
				return true;
			}
			// end loop if last field checked
			if (field == lastfield) {
				break;
			}
		}
		return false;
	}

	/**
	 * Get date from object
	 * @param obj
	 * @return As date
	 */
	private static Date getDate(Object srcObj) {
		if (srcObj instanceof Date) {
			return (Date) srcObj;
		}
		throw new RuntimeException("Can not convert to Date: " + srcObj +", class:" + srcObj.getClass().getName());
	}

	/**
	 * Log info about the verified columns
	 * statistics contains columname=true/false,.... about if the column was verified.
	 */
	public void printStatistics() {
		StringBuffer stat = new StringBuffer();
		for (int i=0; i< columns.size(); i++) {
			String columnName = (String) columns.get(i);
			String jdbcType = JDBCUtils.getJDBCTypeName(getDestJdbcType(columnName));
			if (i !=0 ) {
				stat.append(" ,");
			}
			stat.append(columnName).append("(").append(jdbcType).append(")").append("=").append(columnsVerified.contains(columnName));
		}
		logger.info("Verified columns on table " + tableName + " : " + stat.toString());
	}

}
