/*
 * Copyright by Intland Software
 *
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Intland Software. ("Confidential Information"). You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Intland.
 */
package com.intland.codebeamer.dbcopy;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

/**
 * Factory/registry for sequence copiers. It is a sequence copier itself too, and will delegate to
 * the appropriate sequence copier.
 * @author <a href="mailto:zoltan.luspai@intland.com">Zoltan Luspai</a>
 * @version $Id$
 */
public class SequenceCopiers implements ISequenceCopier{

	private final static Logger logger = Logger.getLogger(SequenceCopiers.class);

	// instance for convinience
	public final static SequenceCopiers INSTANCE = new SequenceCopiers();

	// registry of the sequence copiers, (String)dbName->ISequenceCopier mapping
	private static Map sequenceCopiers = new HashMap();
	static {
		sequenceCopiers.put("MySQL", new MySQLSequenceCopier());
		sequenceCopiers.put("PostgreSQL", new PostgreSQLSequenceCopier());
		sequenceCopiers.put("Apache Derby", new DerbySequenceCopier());
		sequenceCopiers.put("Oracle", new OracleSequenceCopier());
	}

	/**
	 * Determine if the given table has some form of sequence values or auto incrment primary key.
	 * This simply assumes that getNextSequenceValue() will not fail if it has.
	 * @param conn
	 * @param tableName
	 * @return If the table has sequence
	 */
	private static boolean hasSequence(Connection conn, String tableName) {
		try {
			INSTANCE.getNextSequenceValue(conn, tableName);
			return true;
		} catch (SQLException ignored) {
			logger.info("Ignored exception", ignored);
			return false;
		}
	}

	/**
	 * Copy sequence value from one table to the other
	 * @param srcConn The source connection
	 * @param destConn The destination connection
	 * @param tableName The name of the table
	 * @throws SQLException The SQL exception
	 * @return The sequence value copied, -1 if not copied
	 */
	public static int copySequenceValue(Connection srcConn, Connection destConn, String tableName) throws SQLException {
		// Note: I've tried to determine if has sequence by looking at the primary key, but
		// unfortunately this does not work with PostgreSQL !
		/*
		// do not copy if no primary key
		String pkSrc = getPrimaryKeyFieldName(srcConn, tableName);
		String pkDest = getPrimaryKeyFieldName(destConn, tableName);
		String pk = pkSrc;

		if (pk == null) {
			logger.info("Skipping table '" + tableName +"' as it has no primary key!");
			return;
		}
		logger.info("Primary key found for: '" + tableName +"', which is: '" + pk + "'");
		*/
		// check if both the src and dest has sequence keys
		boolean srcHasSequence = hasSequence(srcConn, tableName);
		boolean destHasSequence = hasSequence(destConn, tableName);
		if (!srcHasSequence && !destHasSequence) {
			logger.info("Src and Dest connection reports no sequence on table, skipping:" + tableName);
			return -1;
		}

		int nextIdValue = INSTANCE.getNextSequenceValue(srcConn, tableName);
		// it may happen the next value reports 0 (like for max() functions), this is fixed
		// the 'minimum' next value is 2; means current value is 1
		if (nextIdValue<2) {
			nextIdValue = 2;
		}

		logger.info("Primary key value to be copied:" + nextIdValue);
		INSTANCE.setNextSequenceValue(destConn, tableName, nextIdValue);
		return nextIdValue;
	}

	/**
	 * Query the next sequence value for a table. This is database dependent, because some
	 * databases use Sequences like Oracle and PostgreSQL to generate ids, but mysql uses auto-increment fields
	 *
	 * @param conn The jdbc connection
	 * @param tableName The name of the table
	 * @return The next sequence value
	 * @throws SQLException
	 */
	public int getNextSequenceValue(Connection conn, String tableName) throws SQLException {
		String dbName = JDBCUtils.getDBName(conn);
		logger.debug("Querying next sequence/id value for table '" + tableName + "', db=" + dbName);
		return getSequenceCopier(conn).getNextSequenceValue(conn, tableName);
	}

	/**
	 * Set the next sequence value
	 * @param conn The jdbc connection
	 * @param tableName The name of the table
	 * @throws SQLException
	 */
	public void setNextSequenceValue(Connection conn, String tableName, int nextvalue) throws SQLException {
		String dbName = JDBCUtils.getDBName(conn);
		logger.debug("Setting next sequence/id value for '" + tableName + "', value=" + nextvalue +", db=" + dbName);
		getSequenceCopier(conn).setNextSequenceValue(conn, tableName, nextvalue);
	}

	/**
	 * Get the registered sequence copier for this database
	 * @param conn The connection
	 * @return The sequence copier implementation
	 * @throws SQLException for any sql exception
	 */
	private ISequenceCopier getSequenceCopier (Connection conn) throws SQLException{
		String dbName = JDBCUtils.getDBName(conn);
		ISequenceCopier copier = (ISequenceCopier) sequenceCopiers.get(dbName);
		if (copier == null) {
			throw new RuntimeException("Sequence copier not found for database " + dbName);
		}
		return copier;
	}

	/**
	 * Query the max()+1 as next sequence value for a table
	 * @param conn
	 * @param tableName
	 * @return the next key value would be
	 */
	public static int getMaxOfPrimaryKeyStyleSequenceValue(Connection conn, String tableName) throws SQLException{
		// query the current maximum +1 that's the next sequence value
		String pk = JDBCUtils.getPrimaryKeyFieldName(conn, tableName);
		if (pk == null) {
			throw new SQLException("No primary key found for " + tableName);
		}
		String sql = "SELECT MAX(" + pk +") FROM " + tableName;
		int next = JDBCUtils.querySingleIntValue(conn, sql) + 1;
		if (next < 1) {
			next = 1;
		}
		return next;
	}

	/**
	 * MySQL specific sequence copier.
	 * @author <a href="mailto:zoltan.luspai@intland.com">Zoltan Luspai</a>
	 * @version $Id$
	 */
	static class MySQLSequenceCopier implements ISequenceCopier{

		public int getNextSequenceValue(Connection conn, String tableName) throws SQLException {
			return getMaxOfPrimaryKeyStyleSequenceValue(conn, tableName);
		}

		/**
		 * See: http://dev.mysql.com/doc/refman/5.1/en/example-auto-increment.html
		 */
		public void setNextSequenceValue(Connection conn, String tableName, int nextvalue) throws SQLException {
			String sql = "ALTER TABLE " + tableName +" AUTO_INCREMENT = " + nextvalue;
			JDBCUtils.execUpdateCommand(conn, sql);
		}
	}

	/**
	 * Sequence copier for PostgreSQL, this will use real sequences!
	 * See: http://www.postgresql.org/docs/8.2/interactive/functions-sequence.html
	 *
	 * @author <a href="mailto:zoltan.luspai@intland.com">Zoltan Luspai</a>
	 * @version $Id$
	 */
	static class PostgreSQLSequenceCopier implements ISequenceCopier{

		// how the sequence is named for the table
		private String getSequenceName(String tableName) {
			return tableName + "_s";
		}

		public int getNextSequenceValue(Connection conn, String tableName) throws SQLException {
			// similar to Oracle currval() does not work without selecting the nextval() before
			// would throw:   ERROR:  the_seq.currval is not yet defined in this session
			// so have to use nextval()
			String query = "SELECT NEXTVAL('" + getSequenceName(tableName)+"')";
			return JDBCUtils.querySingleIntValue(conn, query) + 1;
		}

		public void setNextSequenceValue(Connection conn, String tableName, int nextvalue) throws SQLException {
			// only increment sequence value, when it is lower than the desired.
			// this is done to avoid errors like:
			// PSQLException: ERROR: setval: value 14 is out of bounds for sequence "wf_transition_s" (100..9223372036854775807)
			// caused by that the sequence value is lower than the minimum value
			int currentvalue = getNextSequenceValue(conn, tableName);
			if (currentvalue < nextvalue ) {
				String query = "SELECT SETVAL('" + getSequenceName(tableName) +"', " + nextvalue +")";
				// it returns some value, so do a query command
				JDBCUtils.querySingleIntValue(conn, query);
			}
		}

	}

	/**
	 * Derby specific sequence copier.
	 * See: http://db.apache.org/derby/docs/10.2/ref/rrefsqlj81859.html
	 *
	 * @author <a href="mailto:zoltan.luspai@intland.com">Zoltan Luspai</a>
	 * @version $Id$
	 *
	 */
	static class DerbySequenceCopier implements ISequenceCopier{

		public int getNextSequenceValue(Connection conn, String tableName) throws SQLException {
			return getMaxOfPrimaryKeyStyleSequenceValue(conn, tableName);
		}

		public void setNextSequenceValue(Connection conn, String tableName, int nextvalue) throws SQLException {
			String pk = JDBCUtils.getPrimaryKeyFieldName(conn, tableName);
			String sql = "ALTER TABLE " + tableName+ " ALTER COLUMN " + pk + " RESTART WITH " + nextvalue;
			JDBCUtils.execUpdateCommand(conn, sql);
		}
	}

	/**
	 * Oracle specific sequence copier uses real sequences.
	 * See: http://download-uk.oracle.com/docs/cd/B10501_01/server.920/a96540/sql_elements6a.htm#4062
	 *
	 * @author <a href="mailto:zoltan.luspai@intland.com">Zoltan Luspai</a>
	 * @version $Id$
	 *
	 */
	static class OracleSequenceCopier implements ISequenceCopier{

		// how the sequence is named for the table
		private String getSequenceName(String tableName) {
			return tableName + "_s";
		}

		public int getNextSequenceValue(Connection conn, String tableName) throws SQLException {
			// strange: Oracle can not call currval without calling the nextval before
			// it will give an error: ORA-08002: sequence NAME.CURRVAL is not yet defined in this session
			// so have to increment the sequence here!
			String query = "SELECT " + getSequenceName(tableName)+".NEXTVAL FROM DUAL";
			return JDBCUtils.querySingleIntValue(conn, query) + 1;
		}

		public void setNextSequenceValue(Connection conn, String tableName, int nextvalue) throws SQLException {
			// there is no way to directly set the sequence value, but we can increment it
			// see: alter sequence : http://download-uk.oracle.com/docs/cd/B10501_01/server.920/a96540/statements_121a.htm#2052407
			int currval = getNextSequenceValue(conn, tableName);
			int increment = (nextvalue - currval) + 1;
			if (increment > 0) {
				// temporarily set the sequence increment value to higher
				String query = "ALTER SEQUENCE " + getSequenceName(tableName) +" INCREMENT BY " + increment;
				JDBCUtils.execUpdateCommand(conn, query);
				// increment the sequence so it gets the next value
				getNextSequenceValue(conn, tableName);
				// reset the sequence 'increment' value
				query = "ALTER SEQUENCE " + getSequenceName(tableName) +" INCREMENT BY 1 NOCACHE";
				JDBCUtils.execUpdateCommand(conn, query);
			}
		}

	}

}

