/*
 * Copyright by Intland Software
 *
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Intland Software. ("Confidential Information"). You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Intland.
 */
package com.intland.codebeamer.dbcopy;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.NumberFormat;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

/**
 * Class copies one table to some other database.
 *
 * @author <a href="mailto:zoltan.luspai@intland.com">Zoltan Luspai</a>
 * @version $Id$
 */
class TableCopier {

	private final static Logger logger = Logger.getLogger(TableCopier.class);

	private Connection srcConn;
	private Connection destConn;
	private StringBuffer report;
	// List<String> contains the column names must be "quoted" in SQL statements (because they're reserved words)
	private List columnsToQuote;
	private boolean verifyRecords;
	// if first error should abort the copying
	private boolean failFast;
	// if the table has a sequence
	private boolean hasSequence;

	/**
	 * Constructor
	 * @param srcConn Source connection
	 * @param destConn Destination connection
	 * @param columnsToQuote List<String> contains the column names must be "quoted" in SQL statements (because they're reserved words)
	 * @param verifyRecords If the records are verified after insert in the destination database.
	 * @param failFast If first error should abort the copying
	 * @param hasSequence if the table has sequence
	 */
	public TableCopier(Connection srcConn, Connection destConn, List columnsToQuote, boolean verifyRecords, boolean failFast, boolean hasSequence) {
		this.srcConn = srcConn;
		this.destConn = destConn;
		this.columnsToQuote = columnsToQuote;
		this.verifyRecords = verifyRecords;
		this.failFast = failFast;
		this.hasSequence = hasSequence;
	}

	/**
	 * Performs the copy of a single table. Before the copy proceeds, all existing data from the target
	 * table is deleted.
	 * @return If the copy was fully successful. @see getReport() for summary report
	 */
	public boolean copyTable(String tableName) {
		logger.info("Copying table: " + tableName);
		boolean tableCopySuccess = true;

		if (!DatabaseCopier.READ_ONLY) {
			deleteTableData(destConn, tableName);
		}

		Statement readStmt = null;
		PreparedStatement writeStmt = null;
		int insertedRowsCount = 0;
		int totalRowsCount = 0;
		int targetSelectCount = 0;

		String readSQL = null;
		String writeSQL = null;

		NumberFormat numberFormat = NumberFormat.getInstance();

		// workaround for MySQL can use setFetchSize(1), doest not work even with the workaround below!
		boolean fetchRecordsOneByOne = true;
		String primaryKey = null;
		PreparedStatement fetchOneRecordStmt = null;
		if (fetchRecordsOneByOne) {
			// check if fetching one-by-one will work
			try {
				primaryKey = JDBCUtils.getPrimaryKeyFieldName(srcConn, tableName);
			} catch (Exception ex) {
				logger.debug("Exception: can not use fetching-on-by-one, because table does not have a single Primary Key", ex);
				fetchRecordsOneByOne = false;
			}
			if (primaryKey == null) {
				fetchRecordsOneByOne = false;
			}
		}
		logger.debug("Feching records-one-by-one: " + fetchRecordsOneByOne +", on table: " + tableName +",primaryKey:" + primaryKey);

		try {
			final List columns = getTableColumns(tableName);
			// get the jdbc-types for destination connection, this is used for deciding how to read the data
			Map destColumnTypes = JDBCUtils.getColumnTypes(tableName, destConn);

			readSQL = buildReadSQL(columns, tableName);
			writeSQL = buildWriteSQL(columns, tableName);
			logger.debug("Read statement <" + readSQL + "> write statement <" + writeSQL + ">");

			// we'll use batch commit
			destConn.setAutoCommit(!DatabaseCopier.BATCH_COMMIT);

			// to avoid out of memory try fetching one row at a time
			readStmt = srcConn.createStatement(ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_READ_ONLY);
			if (! fetchRecordsOneByOne) {
				setFechSize1(readStmt);
			}
			writeStmt = destConn.prepareStatement(writeSQL);

			String srcDbName = JDBCUtils.getDBName(srcConn);
			String destDbName = JDBCUtils.getDBName(destConn);

			// create a verifier if it is configured so
			RecordVerifier verifier = null;
			if (verifyRecords) {
				if (primaryKey == null) {
					logger.info("Can not verify records on table " + tableName + ", because it has no primary key!");
				} else {
					String fetchOneRecordSQL = readSQL + " WHERE " + primaryKey + " = ?";
					Map srcColumnTypes = JDBCUtils.getColumnTypes(tableName, srcConn);
					verifier = new RecordVerifier(columns, tableName, srcColumnTypes, destConn, fetchOneRecordSQL, primaryKey, destDbName);
				}
			}

			// normal handler for the rows
			RecordCopier recordcopier = new RecordCopier(columns, writeStmt, tableName, destColumnTypes, verifier, srcDbName);
			JDBCUtils.IResultSetHandler rsHandler = recordcopier;

			if (fetchRecordsOneByOne) {
				// fetch one-by-one handler
				// prepare the statement for fetching one record only
				String fetchOneRecordSQL = readSQL + " WHERE " + primaryKey + " = ?";
				fetchOneRecordStmt = srcConn.prepareStatement(fetchOneRecordSQL);
				// replace the handler with the one, which will execute the fetching one-by-one and then delegate to the other handler
				rsHandler = new FetchOneByOneHandler(fetchOneRecordStmt, rsHandler);
			}

			ResultSet readDataResultSet = null;
			long lastPrintoutAt = System.currentTimeMillis();
			int lastRecCount = 0;
			try {
				if (fetchRecordsOneByOne) {
					String fetchIdsSQL = buildReadSQL(Arrays.asList(new String[]{ primaryKey}), tableName);
					logger.debug("Only fetching ids: " + fetchIdsSQL);
					readDataResultSet = readStmt.executeQuery(fetchIdsSQL);
				} else {
					readDataResultSet = readStmt.executeQuery(readSQL);
				}
				while (fetchNextResultSetRecord(readDataResultSet)) {
					try {
						totalRowsCount++;
						insertedRowsCount += ((Integer) rsHandler.handleResults(readDataResultSet)).intValue();

						// batch commit
						if (DatabaseCopier.BATCH_COMMIT && totalRowsCount % 100 ==0) {
							destConn.commit();
						}
					} catch (Exception ex) {
						if (DatabaseCopier.BATCH_COMMIT) {
							try {
								// on exception have to finish transaction on PostgreSQL to be able to insert more records
								// it is committed to get the records so far committed
								destConn.commit();
							} catch (Exception ex2) {
								logger.info("Exception during commit", ex2);
							}
						}
						tableCopySuccess = false;
						logException(tableName, readSQL, writeSQL, recordcopier.getRecordLog(), ex);
					}
					finally {
						writeStmt.clearParameters();
					}

					long elapsedTime = System.currentTimeMillis() - lastPrintoutAt;
					if (logger.isInfoEnabled() && (insertedRowsCount % 1000 == 0 || elapsedTime>50000)) {
						double speed = (totalRowsCount - lastRecCount) / (elapsedTime / 1000.0);
						logger.info("Table: " + tableName + " copied " + numberFormat.format(insertedRowsCount) + " rows.");
						lastRecCount = totalRowsCount;
						lastPrintoutAt = System.currentTimeMillis();
						Runtime.getRuntime().gc();
						logger.debug("Statistics: " + ((int)speed*10)/10 + " records/sec, used memory: " + getUsedMemory()/1024 +"k");
					}
				}
				// print out verifier statistics
				if (verifier != null) {
					verifier.printStatistics();
				}
			} finally {
				// ensure all data commited
				try {
					destConn.commit();
				} catch (Exception ignored) {
					logger.info("Ignored exception on commit in finally:" + ignored.getMessage());
				}

				if (readDataResultSet != null) {
					try {
						readDataResultSet.close();
					} catch (Exception ignored) {
						logger.info("Ignored exception on close of dataset:" + ignored.getMessage());
					}
				}
			}
		}
		catch(Exception ex) {
			tableCopySuccess = false;
			logException(tableName, readSQL, writeSQL, null, ex);
		}
		finally { // cleanup
			closeStatement(readStmt);
			readStmt = null;
			closeStatement(writeStmt);
			writeStmt = null;
			closeStatement(fetchOneRecordStmt);
			fetchOneRecordStmt = null;
		}

		// count the records
		try {
			targetSelectCount = JDBCUtils.querySingleIntValue(destConn, "SELECT COUNT(*) FROM " + tableName);
		} catch (SQLException ex) {
			logger.warn("Exception when counting records in destination table", ex);
		}

		// copy the sequence generator values from one connection to the other
		int sequenceValueCopied = -1;
		if (!DatabaseCopier.READ_ONLY) {
			if (hasSequence) {
				try {
					sequenceValueCopied = SequenceCopiers.copySequenceValue(srcConn, destConn, tableName);

					if (sequenceValueCopied == -1) {
						tableCopySuccess = false;
						logger.error("Sequence copy failed on " + tableName);
					}
				} catch (Exception ex) {
					tableCopySuccess = false;
					logger.error("Failed copying sequence values", ex);
				}
			} else {
				logger.info("Table " + tableName + " has no sequence.");
			}
		}

		if(totalRowsCount != insertedRowsCount || totalRowsCount != targetSelectCount) {
			tableCopySuccess = false;
			logger.warn("Table: " + tableName + " rows (inserted =" + numberFormat.format(insertedRowsCount) + ", number of rows in table:"
					+ numberFormat.format(targetSelectCount) + ") differs from the number of total rows (" + totalRowsCount + ")");
		}
		// collect the final report data
		report = new StringBuffer();
		report.append(DatabaseCopier.padStr(tableName,30, true)).append("\t");
		report.append(DatabaseCopier.padStr(numberFormat.format(totalRowsCount),8, false)).append("\t");
		report.append(DatabaseCopier.padStr(numberFormat.format(targetSelectCount),8, false)).append("\t");
		report.append(DatabaseCopier.padStr(sequenceValueCopied == -1 ? "-" : Integer.toString(sequenceValueCopied), 8, false)).append("\t");
		report.append(tableCopySuccess ? "OK" : "ERRORS");
		report.append("\n");

		logger.info("Table: " + tableName + " " + numberFormat.format(insertedRowsCount) + " rows copied");
		return tableCopySuccess;
	}

	private final static String MYSQL_COMMUNICATION_EXCEPTION_MSG = "Communications link failure due to underlying exception";

	/**
	 * Fetch next record from ResultSet. Essentially same as ResultSet.next(),
	 * but handles MySQL's "com.mysql.jdbc.CommunicationsException: Communications link failure due to underlying exception:"
	 * gracefully.
	 *
	 * @param rs
	 * @return If next record is available
	 * @throws SQLException
	 */
	private boolean fetchNextResultSetRecord(ResultSet rs) throws SQLException {
		try {
			return rs.next();
		} catch (SQLException ex) {
			// workaround for MySQL exception, don't throw exception, it just appears on the end of resultset
			if (ex.getMessage().indexOf(MYSQL_COMMUNICATION_EXCEPTION_MSG) != -1) {
				return false;
			}
			throw ex;
		}
	}

	/**
	 * Get the report string about the success of the table copy.
	 * @return The report string
	 */
	public String getReport() {
		return report == null ? null : report.toString();
	}

	// get the used memory in bytes;
	private long getUsedMemory() {
		return Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
	}

	/**
	 * Log an exception with some extra info
	 * @param tableName
	 * @param readStatement
	 * @param writeStatement
	 * @param recordLog
	 * @param ex
	 */
	private void logException(String tableName, String readStatement, String writeStatement, StringBuffer recordLog, Exception ex) {
		String msg = "Exception while copying table " + tableName;
		logger.error(msg, ex);
		logger.error("Read statement:" + readStatement);
		logger.error("Write statement: " + writeStatement);
		logger.error("Record data: " + recordLog);
		if (failFast) {
			throw new RuntimeException(msg, ex);
		}
	}

	/**
	 * Close one statement, by logging exception during
	 * @param stmt
	 */
	private void closeStatement(Statement stmt) {
		if (stmt != null) {
			try {
				stmt.close();
			} catch (SQLException e) {
				logger.error(e,e);
			}
		}
	}

	/** Deletes all the data from the specified table in the database referenced by the connection. */
	protected void deleteTableData(Connection conn, String tableName) {
		logger.debug("Deleting table data from destination table " + tableName);
		String query = "delete from " + tableName;

		try {
			int numrecords = JDBCUtils.execUpdateCommand(conn, query);
			logger.debug("Deleted " + numrecords +" records.");
		} catch (SQLException ex) {
			logger.error("Error deleting content for table " + tableName + " using query '" + query + "'", ex);
		}
	}

	/** Retrieves the columns of the specified table based on the JDBC database meta data. */
	protected List getTableColumns(String tableName) throws SQLException {
		List columns = JDBCUtils.getTableColumns(tableName, srcConn);

		if(columns.isEmpty()) {
			String dbName = JDBCUtils.getDBName(srcConn);
			logger.warn("No column information could be extracted from the source database of " + dbName +". This is JDBC driver problem.");
			logger.info("Trying to extract column information from the target database.");
			columns = JDBCUtils.getTableColumns(tableName, destConn);

			if(columns.isEmpty()) {
				logger.fatal("Could not extract column information neither from the source nor the target databases.");
				throw new SQLException();
			}
		}

		return columns;
	}

	/** Constructs the SQL statement to be used for reading the data from the source table. */
	protected String buildReadSQL(List columns, String tableName) {
		StringBuffer sql = new StringBuffer("SELECT ");
		sql.append(buildColumnList(tableName + ".", columns, false));
		sql.append(" FROM ");
		sql.append(tableName);
		return sql.toString();
	}

	/**
	 * Build the comma separated column list for a set of columns
	 * @param prefix The prefix to add before column names
	 * @param columns The column names as string
	 * @param quote if quoting the field names are necessary
	 * @return The String contains all column names comma separated
	 */
	private StringBuffer buildColumnList(String prefix, Collection columns, boolean quote) {
		StringBuffer buf = new StringBuffer();
		for(Iterator it = columns.iterator(); it.hasNext(); ) {
			if (buf.length() > 0) {
				buf.append(',');
			}
			// Note: to fix reserved words like "offset" wrap column names in quotes
			if (prefix != null) {
				buf.append(prefix);
			}
			// be 'smart' and quote certain columns which are reserved words in some databases
			String columnName = (String) it.next();
			boolean quote_this_field = quote && columnsToQuote.contains(columnName);
			if (quote_this_field) {
				buf.append('"');
			}
			buf.append(columnName);
			if (quote_this_field) {
				buf.append('"');
			}
		}
		return buf;
	}

	/** Constructs the SQL statement to be used for writing the data to the target table. */
	protected String buildWriteSQL(Collection columns, String tableName) {
		StringBuffer sql = new StringBuffer("INSERT INTO ");
		sql.append(tableName);
		sql.append(" (");
		sql.append(buildColumnList(
					/*tableName + "." PostreSQL does not like the prefix in inserts*/
					null, columns, true));
		sql.append(") VALUES (");

		for(Iterator it = columns.iterator(); it.hasNext(); it.next()) {
			sql.append("?,");
		}

		sql.setLength(sql.length() - 1);
		sql.append(')');

		return sql.toString();
	}

	/**
	 * Set the fetch size to one on the statement
	 * @param readStmt The statement
	 * @throws SQLException
	 */
	private void setFechSize1(Statement readStmt) throws SQLException {
		if (JDBCUtils.getDBName(srcConn).equals(JDBCUtils.DBNAME_MY_SQL)) {
			logger.debug("Using MySQL special for setting fetch size to 1");
			// MySQL driver does not support setFetchSize, except this way:
			// see: http://dev.mysql.com/doc/refman/5.0/en/connector-j-reference-implementation-notes.html
			readStmt.setFetchSize(Integer.MIN_VALUE);
		} else {
			readStmt.setFetchSize(1);
		}
	}

}